"""Holds handlers to manage error tracebacks using."""
import pretty_errors

from dbt_sugar.core.flags import FlagParser


class DbtSugarTracebackManager:
    """Consumes CLI flags (from FlagParser consumer) and sets up traceback pretty formatting."""

    def __init__(self, flags: FlagParser) -> None:
        """Traceback manager constructor.

        Consumes flags from the FlagParser objects and sets up traceback formatting so that we
        can print prettier errors.

        Args:
            flags (FlagParser): [description]
        """
        stack_depth: int = 1
        if flags.verbose:
            stack_depth = 10

        pretty_errors.configure(
            separator_character="*",
            line_number_first=False,
            display_link=True,
            lines_before=5,
            lines_after=2,
            line_color=pretty_errors.RED + "> " + pretty_errors.default_config.line_color,
            code_color="  " + pretty_errors.default_config.line_color,
            truncate_code=True,
            display_locals=True,
            stack_depth=stack_depth,
            trace_lines_before=4,
            trace_lines_after=4,
            display_arrow=True,
        )



"""Main module for dbt-sugar. Sets up CLI arguments and sets up task handlers."""
import argparse
import sys
from typing import List

import pyfiglet
from rich.console import Console

from dbt_sugar.core._version import __version__
from dbt_sugar.core.clients.dbt import DbtProfile, DbtProject
from dbt_sugar.core.config.config import DbtSugarConfig
from dbt_sugar.core.flags import FlagParser
from dbt_sugar.core.logger import GLOBAL_LOGGER as logger
from dbt_sugar.core.logger import log_manager
from dbt_sugar.core.task.audit import AuditTask
from dbt_sugar.core.task.bootstrap import BootstrapTask
from dbt_sugar.core.task.doc import DocumentationTask
from dbt_sugar.core.ui.traceback_manager import DbtSugarTracebackManager
from dbt_sugar.core.utils import check_and_compare_version

console = Console()


def check_and_print_version() -> str:
    """Calls check_and_compare_version and formats a message that works both for main and argparse.

    Returns:
        str: version info message ready for printing
    """
    _, latest_version = check_and_compare_version()
    installed_version_message = f"Installed dbt-sugar version: {__version__}".rjust(40)
    latest_version_message = f"Latest dbt-sugar version: {latest_version}".rjust(40)
    if latest_version:
        return "\n".join([installed_version_message, latest_version_message])
    return installed_version_message


# general parser
parser = argparse.ArgumentParser(
    prog="dbt-sugar",
    formatter_class=argparse.RawTextHelpFormatter,
    description="CLI tool to help users document their dbt models",
    epilog="Select onf of the available sub-commands with --help to find out more about them.",
)

parser.add_argument("-v", "--version", action="version", version=check_and_print_version())

# base sub-parser (sets up args that need to be provided to ALL other sub parsers)
base_subparser = argparse.ArgumentParser(add_help=False)
base_subparser.add_argument(
    "--log-level", help="overrides default log level", type=str, default=str()
)
base_subparser.add_argument(
    "-vv",
    "--verbose",
    help="When provided the length of the tracebacks will not be truncated.",
    action="store_true",
    default=False,
)
base_subparser.add_argument(
    "--syrup",
    help="Name of the syrup confi you wish to use. If left empty dbt-sugar will attempt to read your defaults.",
    type=str,
    default=str(),
)
base_subparser.add_argument(
    "--config-path", help="Full path to config.yml file if not using default."
)
base_subparser.add_argument(
    "--profiles-dir", help="Alternative path to the dbt profiles.yml file.", type=str
)

# Task-specific argument sub parsers
sub_parsers = parser.add_subparsers(title="Available dbt-sugar commands", dest="command")

# DOC task parser
document_sub_parser = sub_parsers.add_parser(
    "doc", parents=[base_subparser], help="Runs documentation and test enforement task."
)
document_sub_parser.set_defaults(cls=DocumentationTask, which="doc")
# TODO: We shouldn't be requiring this if we have a `--model` format as it's considered bad practice
document_sub_parser.add_argument(
    "-m", "--model", help="Name of the dbt model to document", type=str, default=None, required=True
)
document_sub_parser.add_argument(
    "-s",
    "--schema",
    help="Name of the database schema in which the model resides",
    type=str,
    default=None,
)
document_sub_parser.add_argument(
    "--dry-run",
    help="When provided the documentation task will not modify your files",
    action="store_true",
    default=False,
)
document_sub_parser.add_argument(
    "-t",
    "--target",
    help="Which target from the dbt profile to load.",
    type=str,
    default=str(),
)

document_sub_parser.add_argument(
    "--no-ask-tests",
    help="When provided the documentation task will not ask for adding TAGs into the model.",
    action="store_false",
    dest="ask_for_tests",
)

document_sub_parser.add_argument(
    "--ask-tests",
    help="When passed dbt-sugar will ask you if you want to add tests to your models.",
    action="store_true",
    dest="ask_for_tests",
    default=True,
)

document_sub_parser.add_argument(
    "--no-ask-tags",
    help="When provided the documentation task will not ask for adding TAGs into the model.",
    action="store_false",
    dest="ask_for_tags",
)

document_sub_parser.add_argument(
    "--ask-tags",
    help="When passed dbt-sugar will ask you if you want to add tests to your models.",
    action="store_true",
    dest="ask_for_tags",
    default=True,
)

document_sub_parser.add_argument(
    "--use-describe-snowflake",
    help=(
        "When passed dbt doc will use `describe table` the faster but experimental query "
        "to retrive columns from the database."
    ),
    action="store_true",
    default=False,
)

document_sub_parser.add_argument(
    "--preserve-yaml-order",
    help=(
        "When passed, the original order of the model descriptors is preserved, "
        "included any inline or block comments. This means your schema.yml file "
        "will not be sanitised in any way such as alphabetically sorted."
    ),
    action="store_true",
    default=None,
)

# ##### AUDIT Task
audit_sub_parser = sub_parsers.add_parser(
    "audit", parents=[base_subparser], help="Runs audit task."
)
audit_sub_parser.set_defaults(cls=AuditTask, which="audit")
audit_sub_parser.add_argument(
    "-m",
    "--model",
    help="Name of the dbt model to document",
    type=str,
    default=None,
    required=False,
)
audit_sub_parser.add_argument(
    "--bootstrap",
    help=(
        "When passed the audit task will run the bootstrap task "
        "(which creates placeholders for any undocumented model and columns) first"
    ),
    default=False,
    action="store_true",
    required=False,
)
# TODO: see about how we can share arguments between parsers in non inheritance setup
audit_sub_parser.add_argument(
    "--preserve-yaml-order",
    help=(
        "When passed, the original order of the model descriptors is preserved, "
        "included any inline or block comments. This means your schema.yml file "
        "will not be sanitised in any way such as alphabetically sorted."
    ),
    action="store_true",
    default=None,
)


# ##### BOOTSTRAP Task Arg parser
bootstrap_sub_parser = sub_parsers.add_parser(
    "bootstrap",
    parents=[base_subparser],
    help="Runs the bootstrap task, which creates model descriptor files for all your models.",
)
bootstrap_sub_parser.set_defaults(cls=BootstrapTask, which="bootstrap")

# TODO: see about how we can share arguments between parsers
bootstrap_sub_parser.add_argument(
    "--preserve-yaml-order",
    help=(
        "When passed, the original order of the model descriptors is preserved, "
        "included any inline or block comments. This means your schema.yml file "
        "will not be sanitised in any way such as alphabetically sorted."
    ),
    action="store_true",
    default=None,
)


# task handler
def handle(
    parser: argparse.ArgumentParser,
    test_cli_args: List[str] = list(),
) -> int:
    """Task handler factory.

    Args:
        parser (argparse.ArgumentParser): CLI argument parser object.
    """
    flag_parser = FlagParser(parser)
    flag_parser.consume_cli_arguments(test_cli_args=test_cli_args)

    # set up traceback manager fo prettier errors
    DbtSugarTracebackManager(flag_parser)

    sugar_config = DbtSugarConfig(flag_parser)
    sugar_config.load_config()

    dbt_project = DbtProject(
        sugar_config.dbt_project_info.get("name", str()),
        sugar_config.dbt_project_info.get("path", str()),
    )
    dbt_project.read_project()

    dbt_profile = DbtProfile(
        flags=flag_parser,
        profile_name=dbt_project.profile_name,
        target_name=flag_parser.target,
        profiles_dir=flag_parser.profiles_dir,
    )
    dbt_profile.read_profile()

    if flag_parser.log_level == "debug":
        log_manager.set_debug()

    if flag_parser.task == "doc":
        task: DocumentationTask = DocumentationTask(
            flag_parser, dbt_profile, sugar_config, dbt_project._project_dir
        )
        # TODO: We actually need to change the behaviour of DocumentationTask to provide an interactive
        # dry run but for now this allows testing without side effects.
        # the current implementation upsets mypy also.
        if flag_parser.is_dry_run:
            logger.warning("[yellow]Running in --dry-run mode no files will be modified")
            logger.info(f"Would run {task}")
            return 0
        return task.run()

    if flag_parser.task == "audit":
        if flag_parser.run_bootstrap_first:
            logger.info("Running 'bootstrap' task first then auditing your project or model")
            bootstrap_task: BootstrapTask = BootstrapTask(
                flags=flag_parser,
                dbt_path=dbt_project._project_dir,
                sugar_config=sugar_config,
                dbt_profile=dbt_profile,
            )
            bootstrap_task.run()
        audit_task: AuditTask = AuditTask(
            flag_parser,
            dbt_project._project_dir,
            sugar_config=sugar_config,
            dbt_profile=dbt_profile,
        )
        return audit_task.run()

    if flag_parser.task == "bootstrap":
        bootstrap_task = BootstrapTask(
            flags=flag_parser,
            dbt_path=dbt_project._project_dir,
            sugar_config=sugar_config,
            dbt_profile=dbt_profile,
        )
        return bootstrap_task.run()

    raise NotImplementedError(f"{flag_parser.task} is not supported.")


def main(parser: argparse.ArgumentParser = parser, test_cli_args: List[str] = list()) -> int:
    """Just your boring main."""
    exit_code = 0
    _cli_args = test_cli_args or []
    # print version on every run unless doing `--version` which is better handled by argparse
    if "--version" not in sys.argv[1:]:
        version_message = check_and_print_version()
        print(version_message)
        print("\n")
        # print app logo with pyfiglet
        logo_str = str(pyfiglet.figlet_format("dbt-sugar", font="slant"))
        console.print(logo_str, style="blue")
        print("Getting sweetness out of the cupboard 🍬! \n")
    # TODO: Update this when a proper dry-run exists.
    exit_code = handle(parser, _cli_args)  # type: ignore

    if exit_code > 0:
        logger.error("[red]The process you were running did not complete successfully.")
    return exit_code


if __name__ == "__main__":
    exit(main())



"""Audit Task module."""
from pathlib import Path
from typing import Any, Dict, List

from rich import box
from rich.console import Console
from rich.table import Table

from dbt_sugar.core.clients.dbt import DbtProfile
from dbt_sugar.core.clients.yaml_helpers import open_yaml
from dbt_sugar.core.config.config import DbtSugarConfig
from dbt_sugar.core.flags import FlagParser
from dbt_sugar.core.logger import GLOBAL_LOGGER as logger
from dbt_sugar.core.task.base import COLUMN_NOT_DOCUMENTED, BaseTask

console = Console()
NUMBER_COLUMNS_TO_PRINT_PER_ITERACTION = 5


class AuditTask(BaseTask):
    """Audit Task object.

    Holds methods and attrs necessary to audit a model or a dbt project.
    """

    def __init__(
        self,
        flags: FlagParser,
        dbt_path: Path,
        sugar_config: DbtSugarConfig,
        dbt_profile: DbtProfile,
    ) -> None:
        self.dbt_path = dbt_path
        super().__init__(
            flags=flags, dbt_path=self.dbt_path, sugar_config=sugar_config, dbt_profile=dbt_profile
        )
        self.column_update_payload: Dict[str, Dict[str, Any]] = {}
        self._flags = flags
        self.model_name = self._flags.model
        self.model_content: Dict[str, Any] = {}

    def run(self) -> int:
        """Main script to run the command doc"""
        if self.model_name:
            _ = self.is_exluded_model(self.model_name)
            logger.info(f"Running audit of model [bold magenta]{self.model_name}.[/bold magenta]\n")
            path_file, schema_exists, _ = self.find_model_schema_file(self.model_name)
            if not path_file:
                logger.info(f"Could not find {self.model_name} in the project at {self.dbt_path}")
                return 1
            if not schema_exists:
                logger.info("The model is not documented.")
                return 1
            self.model_content = open_yaml(path_file)
            self.derive_model_coverage()
        else:
            logger.info(f"Running audit of dbt project in {self.dbt_path}.\n")
            self.derive_project_coverage()
        return 0

    def derive_model_coverage(self) -> None:
        """Method to get the coverage from a specific model."""
        self.get_model_column_description_coverage()
        self.get_model_test_coverage()

    def derive_project_coverage(self) -> None:
        """Method to get the coverage from a dbt project."""
        self.get_project_column_description_coverage()
        self.get_project_test_coverage()

    def calculate_coverage_percentage(self, misses: int, total: int) -> str:
        """
        Method to calculate the percentage of coverage.

        Args:
            number_failures (int): With the number of failures.
            total (int): With the number of total cases.

        Returns:
            str: with the calculation of the percentage.
        """
        if total == 0:
            return "0.0"

        percentage_failure = round((1 - (misses / total)) * 100, 1)
        return str(percentage_failure)

    def get_model_test_coverage(self) -> None:
        """Method to get the tests coverage from a specific model."""
        # Init variables
        model_number_columns = 0
        model_columns_without_tests = 0
        untested_columns = []

        columns = self.dbt_tests.get(self.model_name)

        if not columns:
            logger.info(
                f"There is no documentation entry for '{self.model_name}' in your schema.yml files. "
                "You might need to run `dbt-sugar doc` first."
            )
            return

        for column in columns:
            model_number_columns += 1
            if len(column["tests"]) == 0:
                model_columns_without_tests += 1
                untested_columns.append(column["name"])

        percentage_not_tested_columns = self.calculate_coverage_percentage(
            misses=model_columns_without_tests, total=model_number_columns
        )

        data = self.print_nicely_the_data(
            data=untested_columns, total=percentage_not_tested_columns
        )

        self.create_table(
            title="Test Coverage", columns=["Untested Columns", r"% coverage"], data=data
        )

    def get_model_column_description_coverage(self) -> None:
        """Method to get the descriptions coverage from a specific model."""
        not_documented_columns = self.get_not_documented_columns(
            schema_content=self.model_content,
            model_name=self.model_name,
        ).keys()

        number_not_documented_columns = len(not_documented_columns)
        number_documented_columns = len(
            self.get_documented_columns(
                schema_content=self.model_content,
                model_name=self.model_name,
            )
        )
        number_columns = number_documented_columns + number_not_documented_columns

        # This means that they are not columns, and we want to skip the printing.
        if number_columns == 0:
            return

        percentage_not_documented_columns = self.calculate_coverage_percentage(
            misses=number_not_documented_columns,
            total=number_columns,
        )
        logger.debug(
            f"percentage_not_documented_columns for '{self.model_name}': {percentage_not_documented_columns}"
        )

        data = self.print_nicely_the_data(
            data=list(not_documented_columns), total=percentage_not_documented_columns
        )

        self.create_table(
            title="Documentation Coverage",
            columns=["Undocumented Columns", r"% coverage"],
            data=data,
        )

    def print_nicely_the_data(self, data: List[str], total: str) -> Dict[str, str]:
        """
        Transforms a list into a dictionary (key: column, value: coverahe) total is at the end.

        Args:
            data (List): list of data to modify.

        Returns:
            Dict[str, str]: with a dictionary with the data as keys and
            the total as a value but only for the last element in the list.
        """
        if data:
            reshaped_data = {column: "" for column in data}
            reshaped_data[""] = ""
            reshaped_data["Total"] = total
            return reshaped_data
        if not data and total == "100.0":
            reshaped_data = {"None": "", "": ""}
            reshaped_data["Total"] = total
            return reshaped_data
        return {}

    def create_table(self, title: str, columns: List[str], data: Dict[str, str]) -> None:
        """
        Method to create a nice table to print the results.

        Args:
            title (str): Title that you want to give to the table.
            columns (List[str]): List of columns that the table is going to have.
            data (Dict[str, str]): with the rows that we want to print.
        """
        table = Table(title=title, box=box.SIMPLE)
        for column in columns:
            table.add_column(column, justify="right", style="bright_yellow", no_wrap=True)

        for model, percentage in data.items():
            table.add_row(model, percentage)

        console = Console()
        console.print(table)

    def get_project_test_coverage(self) -> None:
        """Method to get the model tests coverage per model in a dbt project."""
        print_statistics = {}
        total_number_columns = 0
        number_columns_without_tests = 0

        for model_name in self.dbt_tests.keys():
            columns = self.dbt_tests[model_name]

            model_number_columns = 0
            model_columns_without_tests = 0

            for column in columns:
                total_number_columns += 1
                model_number_columns += 1

                if len(column["tests"]) == 0:
                    number_columns_without_tests += 1
                    model_columns_without_tests += 1

            print_statistics[model_name] = self.calculate_coverage_percentage(
                misses=model_columns_without_tests, total=model_number_columns
            )

        print_statistics[""] = ""
        print_statistics["Total"] = self.calculate_coverage_percentage(
            misses=number_columns_without_tests, total=total_number_columns
        )

        self.create_table(
            title="Test Coverage",
            columns=["Model Name", r"% coverage"],
            data=print_statistics,
        )

    def get_project_column_description_coverage(self) -> None:
        """Method to get the model descriptions coverage per model in a dbt project."""
        print_statistics = {}
        for model_name, path in self.all_dbt_models.items():
            schema_content = open_yaml(path)

            number_documented_columns = len(
                self.get_documented_columns(
                    schema_content=schema_content,
                    model_name=model_name,
                )
            )

            number_not_documented_columns = len(
                self.get_not_documented_columns(
                    schema_content=schema_content,
                    model_name=model_name,
                )
            )

            print_statistics[model_name] = self.calculate_coverage_percentage(
                misses=number_not_documented_columns,
                total=(number_documented_columns + number_not_documented_columns),
            )

        print_statistics[""] = ""
        print_statistics["Total"] = self.get_project_total_test_coverage()

        self.create_table(
            title="Documentation Coverage",
            columns=["Model Name", r"% coverage"],
            data=print_statistics,
        )

    def get_project_total_test_coverage(self) -> str:
        """
        Method to get the descriptions coverage for an entire dbt project.

        Returns:
            str: with global descriptions statistics.
        """
        number_not_documented_columns = 0
        number_of_columns = 0

        for description in self.dbt_definitions.values():
            if description == COLUMN_NOT_DOCUMENTED:
                number_not_documented_columns += 1
            number_of_columns += 1

        return self.calculate_coverage_percentage(
            misses=number_not_documented_columns,
            total=number_of_columns,
        )



"""Holds config for dbt-sugar."""

from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel

from dbt_sugar.core.clients.yaml_helpers import open_yaml
from dbt_sugar.core.exceptions import (
    KnownRegressionError,
    MissingDbtProjects,
    NoSyrupProvided,
    SyrupNotFoundError,
)
from dbt_sugar.core.flags import FlagParser
from dbt_sugar.core.logger import GLOBAL_LOGGER as logger


class DbtProjectsModel(BaseModel):
    """Pydantic validation model for dbt_project dict."""

    name: str
    path: str
    excluded_folders: Optional[Union[List[str], str]] = []
    excluded_models: Optional[Union[List[str], str]] = []


class SyrupModel(BaseModel):
    """Pydantic validation model for syrups dict."""

    name: str
    dbt_projects: List[DbtProjectsModel]
    always_enforce_tests: Optional[bool] = True
    always_add_tags: Optional[bool] = True
    use_describe_snowflake: Optional[bool] = False
    preserve_yaml_order: Optional[bool] = False


class DefaultsModel(BaseModel):
    """Pydantic validation model for defaults dict."""

    syrup: Optional[str]
    target: Optional[str]


class SugarConfigModel(BaseModel):
    """Pydantic validation model for sugar_config dict."""

    defaults: DefaultsModel
    syrups: List[SyrupModel]


class DbtSugarConfig:
    """dbt-sugar configuration class."""

    SUGAR_CONFIG_FILENAME = "sugar_config.yml"
    CLI_OVERRIDE_FLAGS = [
        {"cli_arg_name": "ask_for_tests", "maps_to": "always_enforce_tests"},
        {"cli_arg_name": "ask_for_tags", "maps_to": "always_add_tags"},
        {"cli_arg_name": "use_describe_snowflake", "maps_to": "use_describe_snowflake"},
        {"cli_arg_name": "preserve_yaml_order", "maps_to": "preserve_yaml_order"},
    ]

    def __init__(self, flags: FlagParser, max_dir_upwards_iterations: int = 4) -> None:
        """Constructor for DbtSugarConfig.

        Args:
            flags (FlagParser): consumed flags from FlagParser object.
        """
        self._flags = flags
        self._model_name: str = self._flags.model
        self._task = self._flags.task
        self._config_path = self._flags.config_path
        self._syrup_to_load = flags.syrup
        self._config_file_found_nearby = False
        self._max_folder_iterations = max_dir_upwards_iterations
        self._current_folder = Path.cwd()
        if self._flags.profiles_dir:
            self._current_folder = Path(self._flags.profiles_dir)

        # "externally offered objects"
        self.config_model: SyrupModel

    @property
    def config(self):
        if self.config_model:
            return self._integrate_cli_flags(self.config_model.dict())
        raise AttributeError(f"{type(self).__name__} does not have a parsed config.")

    # ! REGRESSION
    @property
    def dbt_project_info(self):
        """Convenience function to ensure only one dbt project is unders scope

        This was introduced as part of an intentional regresssion because we're not ready
        to support multiple dbt projects yet.
        """
        return self.config.get("dbt_projects", list())[0]

    def load_and_validate_config_yaml(self) -> None:
        yaml_dict = open_yaml(self._config_path)

        # use pydantic to shape and validate
        self._config = SugarConfigModel(**yaml_dict)

    def parse_defaults(self) -> None:
        if self._config.defaults and not self._syrup_to_load:
            self._syrup_to_load = self._config.defaults.dict().get("syrup", str())

    def retain_syrup(self) -> None:
        if not self._syrup_to_load:
            raise NoSyrupProvided(
                "A syrup must be provided either in your config.yml or passed to the CLI. "
                "Run `dbt-sugar --help` for more information."
            )

        for syrup in self._config.syrups:
            syrup_dict = syrup.dict()
            if syrup_dict["name"] == self._syrup_to_load:
                self.config_model = syrup

        if not hasattr(self, "config_model"):
            raise SyrupNotFoundError(
                f"Could not find a syrup named {self._syrup_to_load} in {self._config_path}."
            )

    # TODO: Deprecate this when we lift off and address this regression
    # ! REGRESSION
    def assert_only_one_dbt_project_in_scope(self) -> bool:
        number_of_dbt_projects = len(self.config.get("dbt_projects", list()))
        if number_of_dbt_projects > 1:
            raise KnownRegressionError(
                "dbt-sugar can only support ONE dbt project per sugar. "
                "This limitation will be lifted in the next feature release."
            )
        return True

    def assert_dbt_projects_exist(self) -> bool:
        dbt_projects = self.config["dbt_projects"]

        project_existance = {}
        for project in dbt_projects:
            resolved_dbt_project_path = Path(project["path"]).resolve()
            logger.debug(f"Looking for {project['name']} in {resolved_dbt_project_path}")
            project_existance[project["name"]] = {
                "exists": bool(resolved_dbt_project_path.exists()),
                "path": resolved_dbt_project_path,
            }

        bogus_projects = {
            project: details["path"]
            for project, details in project_existance.items()
            if details["exists"] is False
        }

        # TODO: Maybe we want to revisit this and not have a raise but rather a logger warning and says we'll ignore
        if bogus_projects:
            raise MissingDbtProjects(
                f"The following dbt projects are missing or cannot be found: \n\n{bogus_projects}. \n\n"
                "Check your sugar_config.yml"
            )
        return True

    def locate_config(self) -> None:
        logger.debug(f"Starting config file finding from {self._current_folder}")
        current = self._current_folder
        filename = Path(current).joinpath(self.SUGAR_CONFIG_FILENAME)

        if self._config_path == Path(str()):
            logger.debug("Trying to find sugar_config.yml in current and parent folders")

            folder_iteration = 0
            while folder_iteration < self._max_folder_iterations:
                if filename.exists():
                    sugar_config_dir = filename
                    logger.debug(f"{sugar_config_dir} exists and was retreived.")
                    self._config_path = sugar_config_dir
                    self._config_file_found_nearby = True
                    break
                current = current.parent
                filename = Path(current, self.SUGAR_CONFIG_FILENAME)
                folder_iteration += 1

            else:
                raise FileNotFoundError(
                    f"Unable to find {self.SUGAR_CONFIG_FILENAME} in any nearby "
                    f"directories after {self._max_folder_iterations} iterations upwards."
                )

    def _integrate_cli_flags(self, config_dict: Dict[str, Any]):
        for flag_override_dict in self.CLI_OVERRIDE_FLAGS:
            flag_override = getattr(self._flags, flag_override_dict["cli_arg_name"])
            if flag_override is not None:
                config_dict[flag_override_dict["maps_to"]] = flag_override
        return config_dict

    def load_config(self) -> None:
        self.locate_config()
        self.load_and_validate_config_yaml()
        self.parse_defaults()
        self.retain_syrup()
        self.assert_only_one_dbt_project_in_scope()
        _ = self.assert_dbt_projects_exist()
        logger.debug(f"Config model dict: {self.config_model.dict()}")



"""Contains yaml related utils which might get used in places."""

from pathlib import Path
from typing import Any, Dict

import ruamel.yaml
import yaml
import yamlloader

from dbt_sugar.core.exceptions import YAMLFileEmptyError
from dbt_sugar.core.logger import GLOBAL_LOGGER as logger

# TODO: I probably need to also make sure that we load with the appropriate method
# let's always use ruamel.yaml but in one case we will load with `typ='safe'` and
# in the other we will load with `typ='rt'` the same will be applied to when we save.


def open_yaml(path: Path, preserve_yaml_order: bool = False) -> Dict[str, Any]:
    """Opens a yaml file... Nothing too exciting there.

    Args:
        path (Path): Full filename path pointing to the yaml file we want to open.

    Returns:
        Dict[str, Any]: A python dict containing the content from the yaml file.
    """
    if path.is_file():
        logger.debug(f"Opening: {path}")
        with open(path, "r") as stream:
            if preserve_yaml_order:
                ryaml = ruamel.yaml.YAML(typ="rt")
                yaml_dict = ryaml.load(stream)
            else:
                yaml_dict = yaml.load(stream, Loader=yamlloader.ordereddict.CSafeLoader)
            if yaml_dict:
                return yaml_dict
            raise YAMLFileEmptyError(f"The following file {path.resolve()} seems empty.")
    raise FileNotFoundError(f"File {path.resolve()} was not found.")


def save_yaml(path: Path, data: Dict[str, Any], preserve_yaml_order: bool = False) -> None:
    """Saves a YAML content.

    Args:
        path (Path): Full filename path pointing to the yaml file we want to save.
        data (dict[str, Any]): Data to save in the file.
    """
    # TODO: This is a bit of a hot fix. Ideally we should avoid processing those
    # non-model conatining files earlier in the flow.
    if data.get("models"):
        with open(path, "w") as outfile:
            if preserve_yaml_order:
                ryaml = ruamel.yaml.YAML(typ="rt")
                ryaml.width = 100
                ryaml.dump(data, outfile)
            else:
                yaml.dump(data, outfile, width=100, Dumper=yamlloader.ordereddict.CDumper)



"""Contains all dbt-sugar custom exceptions."""


class DbtSugarException(Exception):
    """Custom Exeption type mainly there to print things in Red."""

    def __init__(self, message: str):  # noqa D107
        # ! TODO: I would like to be able to always colour the message.
        # ! depending what colouring scheme we use we'll have to set this up
        # super().__init__(f"{Fore.RED}{message}"  # example using colorama.
        super().__init__(f"{message}")


class DbtProfileFileMissing(DbtSugarException):
    """Thrown when the `profiles.yml` cannot be found in its expected or provided location."""


class YAMLFileEmptyError(DbtSugarException):
    """Thrown when a yamlfile existed but had nothing in it."""


class ProfileParsingError(DbtSugarException):
    """Thrown when no target entry could be found."""


class SyrupNotFoundError(DbtSugarException):
    """Thrown when a syrup config could not be extracted from the config.yaml."""


class NoSyrupProvided(DbtSugarException):
    """Thrown when neither a default syrup nor a cli-passed syrup can be found."""


class MissingDbtProjects(DbtSugarException):
    """Thrown when one or more in-scope dbt projects could not be found."""


class TargetNameNotProvided(DbtSugarException):
    """Thrown when no `target:` entry is provided in the profiles.yml and not passed on CLI."""


class KnownRegressionError(DbtSugarException):
    """Thrown when we want to warn users of a known regression or limitation that is not implemented."""



"""Flags module containing the FlagParser "Factory"."""
import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import List, Optional


class FlagParser:
    """Sets flags from defaults or by parsing CLI arguments.

    In order to not have to always parse args when testing etc. We set the defaults explicitly here.
    This is a bit strict and not very DRY but it saves from surprises which for now is good.
    """

    def __init__(self, cli_parser: ArgumentParser) -> None:
        """Constructor for FlagParser.

        Holds explicit defaults and consumes parsed flags if asked for it.

        Args:
            cli_parser (ArgumentParser): CLI parser.
        """
        self.cli_parser = cli_parser
        self.model: str = "test_model"
        self.schema: str = ""
        self.log_level: str = "info"
        self.syrup: str = str()
        self.config_path: Path = Path(str())
        self.profiles_dir: Optional[Path] = None
        self.is_dry_run: bool = False
        self.ask_for_tests: bool = True
        self.ask_for_tags: bool = True
        self.target: str = str()
        self.verbose: bool = False
        self.use_describe_snowflake: bool = False

    def consume_cli_arguments(self, test_cli_args: List[str] = list()) -> None:
        _cli_args = test_cli_args or sys.argv[1:]
        self.args = self.cli_parser.parse_args(_cli_args)

        self.task = self.args.command

        # base flags that need to be set no matter what
        if self.args:
            self.log_level = self.args.log_level
            self.verbose = self.args.verbose
            self.syrup = self.args.syrup
            if self.args.profiles_dir:
                self.profiles_dir = Path(self.args.profiles_dir).expanduser()
            if self.args.config_path:
                self.config_path = Path(self.args.config_path).expanduser()

        # task specific args consumption
        if self.task == "audit":
            self.model = self.args.model
            self.run_bootstrap_first = self.args.bootstrap
            self.preserve_yaml_order = self.args.preserve_yaml_order
        elif self.task == "bootstrap":
            self.preserve_yaml_order = self.args.preserve_yaml_order
        elif self.task == "doc":
            self.model = self.args.model
            self.schema = self.args.schema
            self.is_dry_run = self.args.dry_run
            self.target = self.args.target
            # we reverse the flag so that we don't have double negatives later in the code
            self.ask_for_tests = self.args.ask_for_tests
            self.ask_for_tags = self.args.ask_for_tags
            self.use_describe_snowflake = self.args.use_describe_snowflake
            self.preserve_yaml_order = self.args.preserve_yaml_order



"""
Module Snowflake connector.

Module dependent of the base connector.
"""
from typing import Dict, Sequence

import sqlalchemy
from snowflake.sqlalchemy import URL

from dbt_sugar.core.connectors.base import BaseConnector
from dbt_sugar.core.logger import GLOBAL_LOGGER as logger


class SnowflakeConnector(BaseConnector):
    """
    Connection class for Snowflake.

    Child class of base connector.
    """

    def __init__(
        self,
        connection_params: Dict[str, str],
    ) -> None:
        """
        Creates the URL and the Engine for future connections.

        Args:
            connection_params (Dict[str, str]): Dict containing database connection
                parameters and credentials.
        """
        self.connection_url = URL(
            drivername="postgresql+psycopg2",
            user=connection_params.get("user", str()),
            password=connection_params.get("password", str()),
            database=connection_params.get("database", str()),
            account=connection_params.get("account", str()),
            warehouse=connection_params.get("warehouse", str()),
        )
        self.engine = sqlalchemy.create_engine(self.connection_url)

    def get_columns_from_table(
        self, target_table: str, target_schema: str, use_describe: bool = False
    ) -> Sequence[str]:

        # if user wants to use describe (more preformant but with caveat) method
        # we re-implement column describe since snowflake.sqlalchemy is shit.
        if use_describe:
            logger.info(
                "Using [red]experimental[/red] `describe table` feature instead of official "
                "`get_columns` from `snowflake.sqlalchemy`."
            )
            # do some basic escaping to prevent "little bobby drop table" scenario
            target_schema = target_schema.split(";")[0]
            target_table = target_table.split(";")[0]

            connection = self.engine.connect()
            results = connection.execute(f"describe table {target_schema}.{target_table};")
            results = results.fetchall()
            return [row["name"].lower() for row in results]

        # else we just return the base method which will user snowflake.sqlalchemy's impl
        return super().get_columns_from_table(target_table, target_schema)



"""Logger module contains LogManager which sets up file and stream handler + formatting."""

import logging
import re
from pathlib import Path

import logredactor
from rich.logging import RichHandler


class LogManager:
    """Manages the logs formats and levels.

    We have 2 loggers one to stout and one to a logger file. General logger level is DEBUG and
    each handler is set dynamically based on log-level CLI args
    """

    def __init__(
        self,
        log_file_path: Path = Path(Path.cwd(), "dbt_sugar_logs"),
        log_to_console: bool = True,
    ):
        """Log manager constructor. can take and override log path + whether to stout or not.

        Args:
            log_file_path (Path, optional): Custom path to logger file.
                Defaults to Path(Path.cwd(), "dbt_sugar_log").
            log_to_console (bool, optional): When true logs will also be pushed into stout.
                Defaults to True.
        """
        Path(log_file_path).mkdir(parents=True, exist_ok=True)

        log_filename = Path(log_file_path, "dbt_sugar_log.log")
        logger = logging.getLogger("dbt-sugar logger")
        # set the logger to the lowest level (then each handler will have it's level --this ensures
        # that all logging always ends up in the file logger.)
        logger.setLevel(logging.DEBUG)
        # Create handlers
        f_handler = logging.FileHandler(log_filename)
        f_handler.setLevel(logging.DEBUG)

        # Create formatters and add it to handlers
        f_format = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(funcName)s - %(message)s"
        )

        f_handler.setFormatter(f_format)
        # Add handlers to the logger
        logger.addHandler(f_handler)

        # if we want to print the log to console we're going to add a streamhandler
        if log_to_console:
            c_handler = RichHandler(
                rich_tracebacks=True,
                show_level=False,
                markup=True,
                enable_link_path=False,
                show_path=False,
            )
            c_handler.setLevel(logging.INFO)
            logger.addHandler(c_handler)
        redact_patterns = [re.compile(r"(?<=password=).*(?= database)")]
        logger.addFilter(logredactor.RedactingFilter(redact_patterns, default_mask="'*hidden*'"))

        self.logger = logger
        self.f_format = f_format

    def set_debug(self):
        """Set all loggers handlers to debug level."""
        self.logger.setLevel(logging.DEBUG)
        for handler in self.logger.handlers:
            handler.setLevel(logging.DEBUG)


log_manager = LogManager()

GLOBAL_LOGGER = log_manager.logger



"""Utility functions. Things that don't really belong anywhere specific so they end up there..."""

from typing import Optional, Tuple
from urllib.error import URLError

import luddite
from packaging.version import parse as semver_parse

from dbt_sugar.core._version import __version__
from dbt_sugar.core.logger import GLOBAL_LOGGER as logger


def check_and_compare_version(external_version: Optional[str] = str()) -> Tuple[bool, str]:
    """Checks what the currently installed version of dbt-sugar is and compares it to the one on PyPI.

    This requires an internet connection. In the case where this doesn't happen a URLError will
    probably be thrown and in that case we just return False not to cause annoying user experience.

    Args:
        external_version (Optional[str], optional): Mainly for testing purposes. Defaults to str().

    Returns:
        bool: True when dbt-sugar needs an update. False when good.
    """
    try:
        pypi_version: str = luddite.get_version_pypi("dbt-sugar")
        if external_version:
            installed_version = external_version
        else:
            installed_version = __version__

        needs_update = semver_parse(pypi_version) > semver_parse(installed_version)
        if needs_update:
            logger.warning(
                f"[yellow]Looks like you're a bit behind. A newer version of dbt-sugar v{pypi_version} "
                "is available."
            )
        return needs_update, pypi_version

    except URLError:
        return False, str()



"""Holds methods to interact with dbt API (we mostly don't for now because not stable) and objects."""
import os
from pathlib import Path
from typing import Any, Dict, Optional, Union

from pydantic import BaseModel, Field

from dbt_sugar.core.clients.yaml_helpers import open_yaml
from dbt_sugar.core.exceptions import (
    DbtProfileFileMissing,
    ProfileParsingError,
    TargetNameNotProvided,
)
from dbt_sugar.core.flags import FlagParser
from dbt_sugar.core.logger import GLOBAL_LOGGER as logger

DEFAULT_DBT_PROFILE_PATH = Path(os.getenv("DBT_PROFILES_DIR", default=Path.home().joinpath(".dbt")))


class PostgresDbtProfilesModel(BaseModel):
    """Postgres Dbt credentials validation model."""

    type: str
    user: str
    password: str = Field(..., alias="pass")
    database: str = Field(..., alias="dbname")
    target_schema: str = Field(..., alias="schema")
    host: str
    port: int


class SnowflakeDbtProfilesModel(BaseModel):
    """Snowflake Dbt credentials validation model."""

    type: str
    account: str
    user: str
    password: str
    database: str
    target_schema: str = Field(..., alias="schema")
    role: str
    warehouse: str


class DbtProjectModel(BaseModel):
    """Defines pydandic validation schema for a dbt_project.yml file."""

    profile: str


class BaseYamlConfig:
    """Base class object which gets extended by objects which will generally read from yaml configs."""

    def _assert_file_exists(self, dir: Path, filename: str = "profiles.yml") -> bool:
        logger.debug(dir.resolve())
        full_path_to_file = dir / filename
        if full_path_to_file.is_file():
            return True
        else:
            raise DbtProfileFileMissing(f"Could not locate `{filename}` in {dir.resolve()}")


class DbtProject(BaseYamlConfig):
    """Holds parsed dbt project information needed for dbt-sugar such as which db profile to target."""

    DBT_PROJECT_FILENAME: str = "dbt_project.yml"

    def __init__(self, project_name: str, project_dir: Path) -> None:
        """Constructor for DbtProject.

        Given a project name and a project dir it will parse the relevant dbt_project.yml and
        parse information such as `profile` so dbt-sugar knows which database profile entry from
        /.dbt/profiles.yml to use.

        Args:
            project_name (str): Name of the dbt project to read profile from.
            project_dir (Path): Path object the dbt_project.yml to read from.
        """
        self._project_name = project_name
        self._project_dir = project_dir

        # class "outputs"
        self.project: DbtProjectModel
        self.profile_name: str

    @property
    def _dbt_project_filename(self) -> Path:
        logger.debug(f"project_dir: {self._project_dir}")
        return Path(self._project_dir).joinpath(type(self).DBT_PROJECT_FILENAME)

    def read_project(self) -> None:
        _ = self._assert_file_exists(Path(self._project_dir), filename=self.DBT_PROJECT_FILENAME)
        _project_dict = open_yaml(self._dbt_project_filename)

        # pass the dict through pydantic for validation and only getting what we need
        # if the profile is invalid app will crash so no further tests required below.
        logger.debug(f"the project {_project_dict}")
        _project = DbtProjectModel(**_project_dict)
        logger.debug(_project)
        self.project = _project
        self.profile_name = self.project.dict().get("profile", str())

        if not self.profile_name:
            logger.warning(
                f"[yellow]There was no `profile:` entry in {self._dbt_project_filename}. "
                "dbt-sugar will try to find a 'default' profile. This might lead to unexpected"
                "behaviour or an error when no defaulf profile can be found in your dbt profiles.yml"
            )


class DbtProfile(BaseYamlConfig):
    """Holds parsed profile dict from dbt profiles."""

    CLI_OVERRIDE_FLAGS = [{"cli_arg_name": "schema", "maps_to": "target_schema"}]

    def __init__(
        self,
        flags: FlagParser,
        profile_name: str,
        target_name: str,
        profiles_dir: Optional[Path] = None,
    ) -> None:
        """Reads, validates and holds dbt profile info required by dbt-sugar (mainly db creds).

        Args:
            project_name (str): name of the dbt project to read credentials from.
            target_name (str): name of the target entry. This corresponds to what resides below
                "outputs" in the dbt's profile.yml (https://docs.getdbt.com/dbt-cli/configure-your-profile/)
        """
        # attrs parsed from constructor
        self._flags = flags
        self._profile_name = profile_name
        self._target_name = target_name
        self._profiles_dir = profiles_dir

        # attrs populated by class methods
        self.profile: Dict[str, str]

    @property
    def profiles_dir(self):
        if self._profiles_dir:
            return self._profiles_dir
        return DEFAULT_DBT_PROFILE_PATH

    def _get_target_profile(self, profile_dict: Dict[str, Any]) -> Dict[str, Union[str, int]]:
        if self._target_name:
            return profile_dict["outputs"].get(self._target_name)
        self._target_name = profile_dict.get("target", str())
        if self._target_name:
            return profile_dict["outputs"].get(self._target_name)
        else:
            raise TargetNameNotProvided(
                f"No target name provied in {self._profiles_dir} and none provided via "
                "--target in CLI. Cannot figure out appropriate profile information to load."
            )

    def read_profile(self):
        _ = self._assert_file_exists(
            self.profiles_dir
        )  # this will raise so no need to check exists further
        _profile_dict = open_yaml(self.profiles_dir / "profiles.yml")
        _profile_dict = _profile_dict.get(self._profile_name, _profile_dict.get(self._profile_name))
        if _profile_dict:

            # read target name from args or try to get it from the dbt_profile `target:` field.
            _target_profile = self._get_target_profile(profile_dict=_profile_dict)

            if _target_profile:
                _profile_type = _target_profile.get("type")
                # call the right pydantic validator depending on the db type as dbt is not
                # consistent with it's profiles and it's hell to have all the validation in one
                # pydantic model.
                if _profile_type == "snowflake":
                    # uses pydantic to validate profile. It will raise and break app if invalid.
                    _target_profile = SnowflakeDbtProfilesModel(**_target_profile)
                elif _profile_type == "postgres" or "redshift":
                    _target_profile = PostgresDbtProfilesModel(**_target_profile)

                # if we don't manage to read the db type for some reason.
                elif _profile_type is None:
                    raise ProfileParsingError(
                        f"Could not read or find a database type for {self._profile_name} in your dbt "
                        "profiles.yml. Check that this field is not missing."
                    )
                else:
                    raise NotImplementedError(f"{_profile_type} is not implemented yet.")
                logger.debug(_target_profile)
                self.profile = _target_profile.dict()

                # override profile info with potential CLI args
                self._integrate_cli_flags()

            else:
                raise ProfileParsingError(
                    f"Could not find an entry for target: '{self._target_name}', "
                    f"for the '{self._profile_name}' profile in your dbt profiles.yml."
                )

        else:
            raise ProfileParsingError(
                f"Could not find an entry for '{self._profile_name}' in your profiles.yml"
            )

    def _integrate_cli_flags(self) -> None:
        for flag_override_dict in self.CLI_OVERRIDE_FLAGS:
            cli_arg_value = getattr(self._flags, flag_override_dict["cli_arg_name"])
            if cli_arg_value and isinstance(self.profile, dict):
                self.profile[flag_override_dict["maps_to"]] = cli_arg_value
            else:
                logger.debug("No schema passed to CLI will try to read from profile.yml")



"""API definition for Task-like objects."""
import abc
import os
import re
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

from dbt_sugar.core.clients.dbt import DbtProfile
from dbt_sugar.core.clients.yaml_helpers import open_yaml, save_yaml
from dbt_sugar.core.config.config import DbtSugarConfig
from dbt_sugar.core.connectors.postgres_connector import PostgresConnector
from dbt_sugar.core.connectors.redshift_connector import RedshiftConnector
from dbt_sugar.core.connectors.snowflake_connector import SnowflakeConnector
from dbt_sugar.core.flags import FlagParser
from dbt_sugar.core.logger import GLOBAL_LOGGER as logger

COLUMN_NOT_DOCUMENTED = "No description for this column."
MODEL_NOT_DOCUMENTED = "No description for this model."
DEFAULT_EXCLUDED_FOLDERS_PATTERN = r"\/target\/|\/dbt_modules\/"
DEFAULT_EXCLUDED_YML_FILES = r"dbt_project.yml|packages.yml"


DB_CONNECTORS = {
    "postgres": PostgresConnector,
    "snowflake": SnowflakeConnector,
    "redshift": RedshiftConnector,
}


class BaseTask(abc.ABC):
    """Sets up basic API for task-like classes."""

    def __init__(
        self,
        flags: FlagParser,
        dbt_path: Path,
        sugar_config: DbtSugarConfig,
        dbt_profile: DbtProfile,
    ) -> None:
        self.repository_path = dbt_path
        self._sugar_config = sugar_config
        self._flags = flags
        self._dbt_profile = dbt_profile

        # populated by class methods
        self._excluded_folders_from_search_pattern: str = self.setup_paths_exclusion()
        self.all_dbt_models: Dict[str, Path] = {}
        self.dbt_definitions: Dict[str, str] = {}
        self.dbt_tests: Dict[str, List[Dict[str, Any]]] = {}
        self.build_descriptions_dictionary()

    def get_connector(self) -> Union[PostgresConnector, SnowflakeConnector, RedshiftConnector]:
        dbt_credentials = self._dbt_profile.profile
        connector = DB_CONNECTORS.get(dbt_credentials.get("type", ""))
        if not connector:
            raise NotImplementedError(
                f"Connector '{dbt_credentials.get('type')}' is not implemented."
            )

        return connector(dbt_credentials)

    def setup_paths_exclusion(self) -> str:
        """Appends excluded_folders to the default folder exclusion patten."""
        if self._sugar_config.dbt_project_info["excluded_folders"]:
            excluded_folders_from_search_pattern: str = r"\/|\/".join(
                self._sugar_config.dbt_project_info["excluded_folders"]
            )
            return fr"{DEFAULT_EXCLUDED_FOLDERS_PATTERN}|\/{excluded_folders_from_search_pattern}\/"

        else:
            return DEFAULT_EXCLUDED_FOLDERS_PATTERN

    def get_column_description_from_dbt_definitions(self, column_name: str) -> str:
        """Searches for the description of a column in all the descriptions in DBT.

        Args:
            column_name (str): column name to get the description from.

        Returns:
            str: with the description of the column.
        """
        return self.dbt_definitions.get(column_name, COLUMN_NOT_DOCUMENTED)

    def get_documented_columns(
        self, schema_content: Dict[str, Any], model_name: str
    ) -> Dict[str, str]:
        """Method to get the documented columns from a model in a schema.yml.

        Args:
            content (Dict[str, Any]): content of the schema.yml.
            model_name (str): model name to get the columns from.

        Returns:
            Dict[str, str]: with the columns names and descriptions documented.
        """
        documented_columns = {}
        for model in schema_content.get("models", []):
            if model["name"] == model_name:
                for column in model.get("columns", []):
                    if column.get("description", COLUMN_NOT_DOCUMENTED) != COLUMN_NOT_DOCUMENTED:
                        documented_columns[column["name"]] = column["description"]
        return documented_columns

    def column_has_primary_key_tests(
        self, schema_content: Dict[str, Any], model_name: str, column_name: str
    ) -> Optional[bool]:
        """Method to check that the column with the primary key have the unique and not_null tests.

        Args:
            schema_content (Dict[str, Any]): content of the schema.yml.
            model_name (str): model name to check.
            column_name (str): column name with the primary key.

        Returns:
            Optional[bool]: True if the column have unique and not_null tests,
                False if is missing one of them, None if the column don't exists.
        """
        for model in schema_content.get("models", []):
            if model["name"] == model_name:
                for column in model.get("columns", []):
                    if column.get("name", "") == column_name:
                        column_tests = column.get("tests", [])
                        return "unique" in column_tests and "not_null" in column_tests
        return None

    def get_not_documented_columns(
        self, schema_content: Dict[str, Any], model_name: str
    ) -> Dict[str, str]:
        """Method to get the undocumented columns from a model in a schema.yml.

        Args:
            schema_content (Dict[str, Any]): content of the schema.yml.
            model_name (str): model name to get the columns from.

        Returns:
            Dict[str, str]: with the columns names and descriptions undocumented.
        """
        not_documented_columns = {}
        for model in schema_content.get("models", []):
            if model["name"] == model_name:
                for column in model.get("columns", []):
                    if column.get("description", COLUMN_NOT_DOCUMENTED) == COLUMN_NOT_DOCUMENTED:
                        not_documented_columns[column["name"]] = COLUMN_NOT_DOCUMENTED
        return not_documented_columns

    def combine_two_list_without_duplicates(self, list1: List[Any], list2: List[Any]) -> List[Any]:
        """
        Method to combine two list without duplicates.

        Args:
            list1 (List[Any]): First list with any value.
            list2 (List[Any]): Second list with any value.

        Returns:
            List[Any]: with the combine lists.
        """
        if not list1:
            return list2

        for item in list1:
            if item not in list2:
                list2.append(item)
        return list2

    def update_model_description_test_tags(
        self,
        path_file: Path,
        model_name: str,
        dict_column_description_to_update: Dict[str, Dict[str, Any]],
    ):
        """
        Method to update a schema.yml with a Dict of columns names, tests, and tags.

        Args:
            path_file (Path): Path of the schema.yml file to update.
            model_name (str): Name of the model to update.
            dict_column_description_to_update (Dict[str, Dict[str, Any]]): Dict with the column name with
            the description, tags and tests to update.
        """
        content = open_yaml(
            path_file,
            preserve_yaml_order=self._sugar_config.config.get("preserve_yaml_order", False),
        )
        for model in content.get("models", []):
            if model["name"] == model_name:
                for column in model.get("columns", []):
                    column_name = column["name"]
                    if column_name in dict_column_description_to_update:
                        # Update the description
                        description = dict_column_description_to_update[column_name].get(
                            "description"
                        )
                        if description:
                            column["description"] = description

                        # Update the tests without duplicating them.
                        tests = dict_column_description_to_update[column_name].get("tests")
                        if tests:
                            column["tests"] = self.combine_two_list_without_duplicates(
                                column.get("tests", []), tests
                            )

                        # Update the tags without duplicating them.
                        tags = dict_column_description_to_update[column_name].get("tags")
                        if tags:
                            column["tags"] = self.combine_two_list_without_duplicates(
                                column.get("tags", []), tags
                            )
        save_yaml(
            path_file,
            content,
            preserve_yaml_order=self._sugar_config.config.get("preserve_yaml_order", False),
        )

    def update_column_description_from_schema(
        self, path_file: Path, dict_column_description_to_update: Dict[str, Dict[str, Any]]
    ) -> None:
        """Method to update a schema.yml with a Dict of columns names and description.

        Args:
            path_file (Path): Path to the schema.yml file to update the columns descriptions from.
            dict_column_description_to_update (Dict[str, Dict[str, Any]]): Dict with the column name with
            the description to update.
        """
        content = open_yaml(
            path_file,
            preserve_yaml_order=self._sugar_config.config.get("preserve_yaml_order", False),
        )
        for model in content.get("models", []):
            for column in model.get("columns", []):
                column_name = column["name"]
                if column_name in dict_column_description_to_update:
                    new_desctiption = dict_column_description_to_update[column_name].get(
                        "description"
                    )
                    if new_desctiption:
                        column["description"] = new_desctiption
        save_yaml(
            path_file,
            content,
            preserve_yaml_order=self._sugar_config.config.get("preserve_yaml_order", False),
        )

    def update_column_descriptions(
        self, dict_column_description_to_update: Dict[str, Dict[str, Any]]
    ) -> None:
        """Method to update all the schema.ymls from a dbt project with a Dict of columns names and description.

        Args:
            dict_column_description_to_update (Dict[str, Dict[str, Any]]): Dict with the column name with
            the description to update.
        """
        for root, _, files in os.walk(self.repository_path):
            if not re.search(self._excluded_folders_from_search_pattern, root):
                files = [
                    f
                    for f in files
                    if f.lower().endswith(".yml")
                    and not re.search(DEFAULT_EXCLUDED_YML_FILES, f.lower())
                ]
                for file in files:
                    path_file = Path(os.path.join(root, file))
                    self.update_column_description_from_schema(
                        path_file, dict_column_description_to_update
                    )

    def update_test_in_dbt_tests(self, model_name: str, column: Dict[str, Any]) -> None:
        """Update a column tests in the global tests dictionary.

        Args:
            model_name (str): with the model name.
            column (Dict[str, Any]): column information.
        """
        if model_name not in self.dbt_tests:
            self.dbt_tests[model_name] = [
                {"name": column["name"], "tests": column.get("tests", [])}
            ]
        else:
            self.dbt_tests[model_name].append(
                {"name": column["name"], "tests": column.get("tests", [])}
            )

    def update_description_in_dbt_descriptions(
        self, column_name: str, column_description: str
    ) -> None:
        """Update a column description in the global description dictionary.

        Args:
            column_name (str): column name to update.
            column_description (str): column description to update.
        """
        if not column_description:
            column_description = COLUMN_NOT_DOCUMENTED
        self.dbt_definitions[column_name] = column_description

    def remove_excluded_models(self, content: Dict[str, Any]) -> Optional[List[Dict[str, Any]]]:
        """Removes models that are excluded_models from the models dict"""
        models = content.get("models", [])
        # if self._sugar_config.dbt_project_info.get("excluded_models"):
        logger.debug(models)
        if models:
            return [
                model_dict
                for model_dict in models
                if model_dict["name"] not in self._sugar_config.dbt_project_info["excluded_models"]
            ]

        return None

    def read_file(self, filename_path: Path) -> str:
        """
        Method to read a file.

        Args:
            filename_path (Path): full path to the file we want to read.

        Returns:
            str: content of the file.
        """
        content = ""
        if Path(filename_path).exists():
            with open(filename_path, "r") as reader:
                content = reader.read()
        return content

    def load_descriptions_from_a_schema_file(
        self, content: Dict[str, Any], path_schema: Path
    ) -> None:
        """Load the columns descriptions from a schema.yml into the global descriptions cache.

        This cache is used so that we can homogenise descriptions across models and import
        already documented ones.

        Args:
            content (Dict[str, Any]): content of the schema.yaml.
        """
        if not content:
            return
        models = self.remove_excluded_models(content)
        if not models:
            return
        for model in models:
            self.all_dbt_models[model["name"]] = path_schema
            for column in model.get("columns", []):
                column_description = column.get("description", None)
                self.update_description_in_dbt_descriptions(column["name"], column_description)
                self.update_test_in_dbt_tests(model["name"], column)

    def get_file_path_from_sql_model(self, model_name: str) -> Optional[Path]:
        """Get the complete file path from a model name.

        Args:
            model_name (str): with the model name to find.

        Returns:
            Optional[Path]: Path of the SQL file, None if the file doens't exists.
        """
        for root, _, files in os.walk(self.repository_path):
            if not re.search(self._excluded_folders_from_search_pattern, root):
                for file_name in files:
                    file_name = file_name.lower()
                    if file_name == f"{model_name}.sql" and not re.search(
                        DEFAULT_EXCLUDED_YML_FILES, file_name
                    ):
                        return Path(os.path.join(root, file_name))
        return None

    def build_descriptions_dictionary(self) -> None:
        """Load the columns descriptions from all schema files in a dbt project.

        This is purely responsble for building the knowledge of all possible definitions.
        In other words it is independent from the documentation orchestration.
        This happens in the `doc` task
        """
        for root, _, files in os.walk(self.repository_path):
            if not re.search(self._excluded_folders_from_search_pattern, root):
                files = [
                    f
                    for f in files
                    if f.lower().endswith(".yml")
                    and not re.search(DEFAULT_EXCLUDED_YML_FILES, f.lower())
                ]
                for file in files:
                    path_file = Path(os.path.join(root, file))
                    content = open_yaml(
                        path_file,
                        preserve_yaml_order=self._sugar_config.config.get(
                            "preserve_yaml_order", False
                        ),
                    )
                    logger.debug(path_file)
                    if content.get("models"):
                        self.load_descriptions_from_a_schema_file(content, path_file)

    def is_model_in_schema_content(self, content, model_name) -> bool:
        """Method to check if a model exists in a schema.yaml content.

        Args:
            content (Dict[str, Any]): content of the schema.yaml.
            model_name (str): model name to search.

        Returns:
            boolean: is true if the model is present in the schema.yaml.
        """
        if not content:
            return False

        return any(model["name"] == model_name for model in content.get("models", []))

    def find_model_schema_file(self, model_name: str) -> Tuple[Optional[Path], bool, bool]:
        for root, _, files in os.walk(self.repository_path):
            if not re.search(self._excluded_folders_from_search_pattern, root):
                schema_file_path = None
                model_file_found = False
                schema_file_exists = False
                is_already_documented = False
                for file in files:
                    # check the model file exists and if it does return the path
                    # of the schema.yml it's in.
                    if file == f"{model_name}.sql":
                        model_file_found = True
                        logger.debug(f"Found sql file for '{model_name}'")
                        schema_file_path = self.all_dbt_models.get(model_name, None)
                    # if it's not in a schema file, then it's not documented and we
                    # need to create a schema.yml "dummy" to place it in.
                    if not schema_file_path and model_file_found:
                        logger.debug(
                            f"'{model_name}' was not contained in a schema file. Creating one at {root}"
                        )
                        schema_file_path = Path(os.path.join(root, "schema.yml"))
                        # check whether there is a schema file already present
                        schema_file_exists = False
                        if schema_file_path.exists():
                            schema_file_exists = True
                        return schema_file_path, schema_file_exists, is_already_documented

                    if schema_file_path and model_file_found:
                        logger.debug(
                            f"'{model_name}' found in '{schema_file_path}' we'll update entry."
                        )
                        is_already_documented = True
                        schema_file_exists = True
                        return schema_file_path, schema_file_exists, is_already_documented
        return None, False, False

    def is_exluded_model(self, model_name: str) -> bool:
        if model_name in self._sugar_config.dbt_project_info.get("excluded_models", []):
            raise ValueError(
                f"You decided to exclude '{model_name}' from dbt-sugar's scope. "
                f"You run `{self._flags.task}` on it you will need to remove "
                "it from the excluded_models list in the sugar_config.yml"
            )
        return True

    @abc.abstractmethod
    def run(self) -> int:
        """Orchestrator method that calls all the needed stuff to run a documentation task."""
        ...



"""Document Task module."""
import copy
import re
import subprocess
from collections import OrderedDict
from pathlib import Path
from shlex import quote
from typing import Any, Dict, List, Mapping, Optional, Sequence

from rich.console import Console
from rich.progress import Progress, SpinnerColumn

from dbt_sugar.core.clients.dbt import DbtProfile
from dbt_sugar.core.clients.yaml_helpers import open_yaml, save_yaml
from dbt_sugar.core.config.config import DbtSugarConfig
from dbt_sugar.core.connectors.postgres_connector import PostgresConnector
from dbt_sugar.core.connectors.redshift_connector import RedshiftConnector
from dbt_sugar.core.connectors.snowflake_connector import SnowflakeConnector
from dbt_sugar.core.flags import FlagParser
from dbt_sugar.core.logger import GLOBAL_LOGGER as logger
from dbt_sugar.core.task.base import MODEL_NOT_DOCUMENTED, BaseTask
from dbt_sugar.core.ui.cli_ui import UserInputCollector

console = Console()
NUMBER_COLUMNS_TO_PRINT_PER_ITERACTION = 5

DB_CONNECTORS = {
    "postgres": PostgresConnector,
    "snowflake": SnowflakeConnector,
    "redshift": RedshiftConnector,
}
PRIMARY_KEYS_TESTS = ["unique", "not_null"]


class DocumentationTask(BaseTask):
    """Documentation Task object.

    Holds methods and attrs necessary to orchestrate a model documentation task.
    """

    def __init__(
        self, flags: FlagParser, dbt_profile: DbtProfile, config: DbtSugarConfig, dbt_path: Path
    ) -> None:
        super().__init__(
            flags=flags, dbt_path=dbt_path, sugar_config=config, dbt_profile=dbt_profile
        )
        self.column_update_payload: Dict[str, Dict[str, Any]] = {}
        self._flags = flags
        self._dbt_profile = dbt_profile
        self._sugar_config = config
        self.dbt_path = dbt_path

    def run(self) -> int:
        """Main script to run the command doc"""
        columns_sql = []

        model = self._flags.model
        schema = self._dbt_profile.profile.get("target_schema", "")

        dbt_credentials = self._dbt_profile.profile
        connector = DB_CONNECTORS.get(dbt_credentials.get("type", ""))
        if not connector:
            raise NotImplementedError(
                f"Connector '{dbt_credentials.get('type')}' is not implemented."
            )

        self.connector = connector(dbt_credentials)

        # exit early if model is in the excluded_models list
        _ = self.is_exluded_model(model)
        columns_sql = self.connector.get_columns_from_table(
            model, schema, self._sugar_config.config.get("use_describe_snowflake", False)
        )
        if columns_sql:
            return self.orchestrate_model_documentation(schema, model, columns_sql)
        return 1

    def update_model_description(
        self, content: Dict[str, Any], model_name: str, is_already_documented: bool = False
    ) -> Dict[str, Any]:
        """Updates the model description from a schema.yaml.

        Args:
            content (Dict[str, Any]): Schema.yml Content.
            model_name (str): Name of the model for which the description will be changed.

        Returns:
            Dict[str, Any]: Schema.yml content updated.
        """
        message = f"Do you want to write a description for {model_name}"
        if is_already_documented:
            message = f"Do you want to change the model description of {model_name}"
        model_doc_payload: List[Mapping[str, Any]] = [
            {
                "type": "confirm",
                "name": "wants_to_document_model",
                "message": message,
                "default": True,
            },
            {
                "type": "text",
                "name": "model_description",
                "message": "Please write down your description:",
            },
        ]
        user_input = UserInputCollector("model", model_doc_payload).collect()
        if user_input.get("model_description", None):
            for model in content.get("models", []):
                if model["name"] == model_name:
                    model["description"] = user_input["model_description"]
        return content

    def move_name_and_description_to_first_position(self, model: Dict[str, Any]):
        """
        Move keys name and description to the beginning of the dictionary.

        Args:
            model (Dict[str, Any]): Name of the model to document.

        Returns:
            Dict[str, Any]: with the keys name and description in the begining of the dictionary.
        """
        # DEPRECATION: Drop ordered dict when dropping python 3.6 support
        ordered_dict = OrderedDict(model)
        if ordered_dict.get("description"):
            ordered_dict.move_to_end("description", last=False)
        ordered_dict.move_to_end("name", last=False)
        return ordered_dict

    def order_schema_yml(self, content_yml: Dict[str, Any]):
        """
        Order the content from a schema.yml. Will order:

            - The schema.yml content in alphabetical order.
            - The models in alphabetical order.
            - The columns from each model in alphabetical order.

        Args:
            content_yml (Dict[str, Any]): schema yml content to order.
        Returns:
            content_yml (Dict[str, Any]): schema yml content with the content ordered.
        """
        for i, model in enumerate(content_yml.get("models", {})):
            # Ensure model name and description outer keys are in first position
            content_yml["models"][i] = self.move_name_and_description_to_first_position(
                content_yml["models"][i]
            )

            # Sort columns names in alphabetical order inside the model.
            if model.get("columns", None):
                content_yml["models"][i]["columns"] = sorted(
                    model["columns"],
                    key=lambda k: k["name"].lower(),
                )
                # ensure name and description are in first position for each column entry.
                content_yml["models"][i]["columns"] = [
                    self.move_name_and_description_to_first_position(column_dict)
                    for column_dict in content_yml["models"][i]["columns"]
                ]
        # Sorting models names in alphabetical order.
        content_yml["models"] = sorted(content_yml["models"], key=lambda k: k["name"].lower())
        return content_yml

    def get_primary_key_from_sql(self, sql_file_path: Path) -> Optional[str]:
        """
        Gets the primary key info from a dbt model's config block.

        Args:
            sql_file_path (Path): full path including file name of a dbt .sql file.

        Returns:
            Optional[str]: name of the primary key column. None if no primary key is specified in the config block.
        """
        sql_content = self.read_file(sql_file_path)
        unique_key = re.search(r"unique_key[^\S]*=[^\S]*\'([a-z_]+)\'", sql_content)
        if unique_key:
            return unique_key.group(1)
        return None

    def add_primary_key_tests(
        self,
        schema_content: Dict[str, Any],
        model_name: str,
    ) -> None:
        """
        Adds the primary key tests (unique, not_null) to the primary key column.

        Args:
            schema_content (Dict[str, Any]): content of the schema.yml.
            model_name (str): Name of the model on which to add primary key tests.
        """
        model_file_path = self.get_file_path_from_sql_model(model_name=model_name)
        if not model_file_path:
            return

        primary_key_column = self.get_primary_key_from_sql(model_file_path)
        if not primary_key_column:
            logger.info(
                f"[bold][yellow]Pro Tip:[/yellow][/bold] The model '{model_name}' does not "
                "have a primary key specified in the "
                "`{{config()}}` block.\nIf you add it, "
                "dbt-sugar can [bold]automatically[/bold] enforce 'unique' and 'not_null' "
                "tests on this column.\n"
                "[bold]New in dbt-sugar 0.1.0!\n"
                # DEPRECATION: remove this novely message in the next feature realease.
            )
            return

        has_primary_key_tests = self.column_has_primary_key_tests(
            schema_content=schema_content, model_name=model_name, column_name=primary_key_column
        )

        if has_primary_key_tests is False:
            logger.info(
                f"""\nAutomatic Process: We have detected that column '{primary_key_column}'
                is a primary key, 'unique' and 'not_null' tests will be added for you.\n"""
            )
            if primary_key_column not in self.column_update_payload.keys():
                self.column_update_payload[primary_key_column] = {"tests": PRIMARY_KEYS_TESTS}
            else:
                tests = self.column_update_payload[primary_key_column].get("tests", [])
                self.column_update_payload[primary_key_column][
                    "tests"
                ] = self.combine_two_list_without_duplicates(PRIMARY_KEYS_TESTS, tests)

    def orchestrate_model_documentation(
        self, schema: str, model_name: str, columns_sql: List[str]
    ) -> int:
        """
        Orchestrator to fully document a model will:

            - Create or update the columns found the database table into the schema.yml.
            - Gives the user the posibility to change the model description.
            - Gives the user the posibility to document the undocumented columns.
            - Updates all the new columns definitions in all dbt.

        Args:
            model_name (str): Name of the model to document.
            columns_sql (List[str]): Columns names that the model have in the database.

        Returns:
            int: with the status of the execution. 1 for fail, and 0 for ok!
        """
        content = None
        schema_file_path, schema_exists, is_already_documented = self.find_model_schema_file(
            model_name
        )

        if not schema_file_path:
            raise FileNotFoundError(
                f"Model: '{model_name}' could not be found in your dbt project."
            )
        if schema_exists:
            content = open_yaml(
                schema_file_path,
                preserve_yaml_order=self._sugar_config.config.get("preserve_yaml_order", False),
            )

        content = self.create_or_update_model_entry(
            is_already_documented, content, model_name, columns_sql
        )
        try:
            content = self.update_model_description(content, model_name, is_already_documented)

            not_documented_columns = self.get_not_documented_columns(content, model_name)
            self.document_columns(not_documented_columns, "undocumented_columns")

            documented_columns = self.get_documented_columns(content, model_name)
            self.document_columns(documented_columns, "documented_columns")
        except KeyboardInterrupt:
            logger.info("The user has exited the doc task, all changes have been discarded.")
            return 0

        if self._sugar_config.config.get("preserve_yaml_order"):
            save_yaml(schema_file_path, content, preserve_yaml_order=True)
        else:
            save_yaml(schema_file_path, self.order_schema_yml(content))
        self.add_primary_key_tests(schema_content=content, model_name=model_name)

        # The copy is here because it was modifying the tests.
        self.update_model_description_test_tags(
            schema_file_path, model_name, copy.deepcopy(self.column_update_payload)
        )
        self.check_tests(schema_file_path, model_name)
        # Method to update the descriptions in all the schemas.yml
        self.update_column_descriptions(self.column_update_payload)

        return 0

    def delete_failed_tests_from_schema(
        self, path_file: Path, model_name: str, tests_to_delete: Dict[str, List[str]]
    ):
        """
        Method to delete the failing tests from the schema.yml.

        Args:
            path_file (Path): Path of the schema.yml file to update.
            model_name (str): Name of the model to document.
            tests_to_delete (Dict[str, List[str]]): with the tests that have failed.
        """
        content = open_yaml(
            path_file,
            preserve_yaml_order=self._sugar_config.config.get("preserve_yaml_order", False),
        )
        for model in content["models"]:
            if model["name"] == model_name:
                for column in model.get("columns", []):
                    tests_to_delete_from_column = tests_to_delete.get(column["name"], [])
                    tests_from_column = column.get("tests", [])
                    tests_pass = [
                        x for x in tests_from_column if x not in tests_to_delete_from_column
                    ]
                    if not tests_pass and tests_from_column:
                        del column["tests"]
                    elif tests_pass:
                        column["tests"] = tests_pass
        save_yaml(
            path_file,
            content,
            preserve_yaml_order=self._sugar_config.config.get("preserve_yaml_order", False),
        )

    def check_tests(self, path_file: Path, model_name: str) -> None:
        """
        Method to run and add test into a schema.yml, this method will:

        Run the tests and if they have been successful it will add them into the schema.yml.

        Args:
            path_file (Path): Path of the schema.yml file to update.
            model_name (str): Name of the model to document.
        """
        with Progress(
            "[progress.description]{task.description}",
            SpinnerColumn(spinner_name="pong"),
            transient=True,
        ) as progress:
            test_checking_task = progress.add_task(
                "[bold]checking your tests via dbt... [/bold]",
                total=len(self.column_update_payload.keys()),
            )
            dbt_command = (
                f"dbt test --models {quote(model_name)} --project-dir {self.dbt_path}".split()
            )
            dbt_result_command = subprocess.run(dbt_command, capture_output=True, text=True).stdout
            tests_to_delete: Dict[str, List[str]] = {}

            if "Compilation Error" in dbt_result_command:
                logger.info(
                    "dbt encountered a compilation error in one or more of your custom tests.\n"
                    "Not able to check if the tests that you have added have PASSED.\n"
                    f"This is what dbt's compilation error says:\n{dbt_result_command}"
                )

            for column in self.column_update_payload.keys():
                tests = self.column_update_payload[column].get("tests", [])
                for test in tests:
                    test_name = test if isinstance(test, str) else list(test.keys())[0]
                    test_passed_pattern = f"PASS {test_name}_{model_name}_{column}"
                    if re.search(test_passed_pattern, dbt_result_command):
                        has_passed = True
                    else:
                        has_passed = False
                        tests_to_delete[column] = tests_to_delete.get(column, []) + [test]
                    test_checking_message = self._generate_test_success_message(
                        test_name, column, has_passed
                    )
                    progress.console.log(test_checking_message)
                progress.advance(test_checking_task)
        if tests_to_delete:
            self.delete_failed_tests_from_schema(path_file, model_name, tests_to_delete)

    @staticmethod
    def _generate_test_success_message(test_name: str, column_name: str, has_passed: bool):
        if has_passed:
            return f"The [bold]{test_name}[/bold] test on '{column_name}' [green]PASSED"
        return (
            f"The [bold]{test_name}[/bold] test on '{column_name}' [red]FAILED[/red]. \n\t└[bold]It will not be added "
            "to your schema.yml.[/bold]"
        )

    def document_columns(
        self, columns: Dict[str, str], question_type: str = "undocumented_columns"
    ) -> None:
        """
        Method to document the columns from a model.

        Will ask the user which columns they want to document and collect the new definition.

        Args:
            columns (Dict[str, str]): Dict of columns with the column name as the key
            and the column description to populate schema.yml as the value.
        """
        allowed_question_types_map = {
            "undocumented_columns": "undocumented columns",
            "documented_columns": "documented columns",
        }
        assert (
            question_type in allowed_question_types_map
        ), f"question_type must be one of those: {list(allowed_question_types_map.keys())}"

        # set up pagination messaging
        columns_names = list(columns.keys())
        number_of_colums_to_document = len(columns_names)
        is_paginated = number_of_colums_to_document > NUMBER_COLUMNS_TO_PRINT_PER_ITERACTION
        if is_paginated:
            logger.info(
                f"There are {number_of_colums_to_document} columns to document in total we will show them to you "
                f"{NUMBER_COLUMNS_TO_PRINT_PER_ITERACTION} at a time."
            )

        # go through columns to document
        for i in range(0, number_of_colums_to_document, NUMBER_COLUMNS_TO_PRINT_PER_ITERACTION):
            final_index = i + NUMBER_COLUMNS_TO_PRINT_PER_ITERACTION
            is_first_page = final_index <= NUMBER_COLUMNS_TO_PRINT_PER_ITERACTION

            choices_undocumented = columns_names[i:final_index]
            choices_documented = {}
            # Feed the current description into the choices messahe.
            if question_type == "documented_columns":
                choices_documented = {key: columns[key] for key in choices_undocumented}
            choices = choices_documented or choices_undocumented

            payload: List[Mapping[str, Any]] = [
                {
                    "type": "checkbox",
                    "name": "cols_to_document",
                    "choices": choices,
                    "message": "Select the columns you want to document.",
                }
            ]
            user_input = UserInputCollector(
                question_type,
                payload,
                ask_for_tests=self._sugar_config.config["always_enforce_tests"],
                ask_for_tags=self._sugar_config.config["always_add_tags"],
                is_paginated=is_paginated,
                is_first_page=is_first_page,
            ).collect()
            self.column_update_payload.update(user_input)

    def update_model(
        self, content: Dict[str, Any], model_name: str, columns_on_db: Sequence[str]
    ) -> Dict[str, Any]:
        """Method to update the columns from a model in a schema.yaml content.

        Args:
            content (Dict[str, Any]): content of the schema.yaml.
            model_name (str): model name to update.
            columns_on_db (List[str]): List of columns that the model has in the database.

        Returns:
            Dict[str, Any]: with the content of the schema.yml with the model updated.
        """
        logger.info(f"The model '{model_name}' already exists, updating its documentation.")
        for model in content.get("models", []):
            if model["name"] == model_name:
                for column in columns_on_db:
                    # Check whether the model has columns already documented in schema yaml
                    if not model.get("columns", None):
                        model["columns"] = []

                    columns = model.get("columns", [])
                    columns_names = [column["name"] for column in columns]
                    if column not in columns_names:
                        description = self.get_column_description_from_dbt_definitions(column)
                        logger.info(f"Updating column '{column.lower()}'")
                        columns.append({"name": column, "description": description})
        return content

    def create_new_model(
        self, content: Optional[Dict[str, Any]], model_name: str, columns_sql: Sequence[str]
    ) -> Dict[str, Any]:
        """Method to create a new model in a schema.yaml content.

        Args:
            content (Dict[str, Any]): content of the schema.yaml.
            model_name (str): Name of the model for which to create entry in schema.yml.
            columns_sql (List[str]): List of columns that the model have in the database.

        Returns:
            Dict[str, Any]: with the content of the schema.yml with the model created.
        """
        logger.info(f"The model '{model_name}' has not been documented yet. Creating a new entry.")
        columns = []
        for column_sql in columns_sql:
            description = self.get_column_description_from_dbt_definitions(column_sql)
            columns.append({"name": column_sql, "description": description})
        model = {
            "name": model_name,
            "description": MODEL_NOT_DOCUMENTED,
            "columns": columns,
        }
        if not content:
            content = {"version": 2, "models": [model]}
        else:
            content["models"].append(model)
        return content

    def create_or_update_model_entry(
        self,
        is_already_documented: bool,
        content: Optional[Dict[str, Any]],
        model_name: str,
        columns_sql: Sequence[str],
    ) -> Dict[str, Any]:
        """Method to update/create a model entry in the schema.yml.

        Args:
            model_name (str): Name of the model for which to create or update entry in schema.yml.
            columns_sql (List[str]): List of columns names found in the database for this model.
        """
        if is_already_documented and content:
            content = self.update_model(content, model_name, columns_sql)
        else:
            content = self.create_new_model(content, model_name, columns_sql)
        return content



"""
Module Postgres connector.

Module dependent of the base connector.
"""
from typing import Dict

import sqlalchemy

from dbt_sugar.core.connectors.base import BaseConnector


class PostgresConnector(BaseConnector):
    """
    Connection class for Postgres.

    Child class of base connector.
    """

    def __init__(
        self,
        connection_params: Dict[str, str],
    ) -> None:
        """
        Creates the URL and the Engine for future connections.

        Args:
            connection_params (Dict[str, str]): Dict containing database connection
                parameters and credentials.
        """
        self.connection_url = sqlalchemy.engine.url.URL(
            drivername="postgresql+psycopg2",
            host=connection_params.get("host", str()),
            username=connection_params.get("user", str()),
            password=connection_params.get("password", str()),
            database=connection_params.get("database", str()),
            port=connection_params.get("port", str()),
        )
        self.engine = sqlalchemy.create_engine(self.connection_url)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['dbt_sugar',
 'dbt_sugar.core',
 'dbt_sugar.core.clients',
 'dbt_sugar.core.config',
 'dbt_sugar.core.connectors',
 'dbt_sugar.core.task',
 'dbt_sugar.core.ui']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'SQLAlchemy>=1.3.0,<1.4.0',
 'click>=7.1.2,<8.0.0',
 'logredactor>=0.0.1,<0.0.2',
 'luddite>=1.0.1,<2.0.0',
 'packaging>=20.8,<21.0',
 'pretty-errors>=1.2.19,<2.0.0',
 'psycopg2>=2.8.6,<3.0.0',
 'pydantic>=1.8,<2.0',
 'pyfiglet>=0.8.post1,<0.9',
 'questionary>=1.9.0,<2.0.0',
 'rich>=9.13,<11.0',
 'ruamel.yaml>=0.17.10,<0.18.0',
 'snowflake-sqlalchemy==1.2.4',
 'sqlalchemy-redshift>=0.8.2,<0.9.0',
 'yamlloader>=1.0.0,<2.0.0']

entry_points = \
{'console_scripts': ['dbt-sugar = dbt_sugar.core.main:main']}

setup_kwargs = {
    'name': 'dbt-sugar',
    'version': '0.1.0',
    'description': 'A sweet CLI tool to help dbt users enforce documentation and testing on their dbt projects.',
    'long_description': "![output-onlineasciitools](https://user-images.githubusercontent.com/4304794/116912589-b806ac80-ac48-11eb-9094-813a0502eb94.png)\n\n[![PyPI version](https://badge.fury.io/py/dbt-sugar.svg)](https://badge.fury.io/py/dbt-sugar)\n\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/ambv/black)\n[![Checked with mypy](http://www.mypy-lang.org/static/mypy_badge.svg)](http://mypy-lang.org/)\n[![Imports: isort](https://img.shields.io/badge/%20imports-isort-%231674b1?style=flat&labelColor=ef8336)](https://pycqa.github.io/isort/)\n![python](https://img.shields.io/badge/python-3.6%20%7C%203.7%20%7C%203.8%20%7C%203.9-blue)\n\n![Build](https://github.com/bitpicky/dbt-sugar/actions/workflows/main_ci.yml/badge.svg)\n[![pre-commit.ci status](https://results.pre-commit.ci/badge/github/bitpicky/dbt-sugar/main.svg)](https://results.pre-commit.ci/latest/github/bitpicky/dbt-sugar/main)\n[![codecov](https://codecov.io/gh/bitpicky/dbt-sugar/branch/main/graph/badge.svg?token=JB0E0LZDW1)](https://codecov.io/gh/bitpicky/dbt-sugar)\n[![Maintainability](https://api.codeclimate.com/v1/badges/1e6a887de605ef8e0eca/maintainability)](https://codeclimate.com/github/bitpicky/dbt-sugar/maintainability)\n\n[![Downloads](https://pepy.tech/badge/dbt-sugar)](https://pepy.tech/project/dbt-sugar)\n[![Discord](https://img.shields.io/discord/752101657218908281?label=discord)](https://discord.gg/cQB49ejbCA)\n\nDocumentation on [gitbook](https://bitpicky.gitbook.io/dbt-sugar/)\n\nIf you want to help out and join the party feel free to:\n\n- [create an issue](https://github.com/bitpicky/dbt-sugar/issues) or more fun\n- join us on [Discord](https://discord.gg/cQB49ejbCA) :sparkles: and\n- check the [ROADMAP.md](ROADMAP.md) to see what's in the plans.\n\n# dbt-sugar :candy:\n\n:warning: **dbt-sugar is pretty alpha, don't use on your prod models unless you have tested it in a safe place before.**\n\n## What is dbt-sugar?\n\ndbt-sugar is a CLI tool that allows users of [dbt](https://www.getdbt.com/) to have fun and ease performing actions around dbt models such as:\n\n- documentation\n- test enforcement\n  and probably more in a not too distant future.\n\n  We keep track of our progress in the [projects](https://github.com/bitpicky/dbt-sugar/projects) section.\n",
    'author': 'Bastien Boutonnet',
    'author_email': 'bastien.b1@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bitpicky/dbt-sugar',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.3,<3.10',
}


setup(**setup_kwargs)



"""User Input Collector API."""

import copy
from typing import Any, Dict, List, Mapping, Optional, Sequence, Union, cast

import click
import questionary
import yaml
from pydantic import BaseModel, validator

DESCRIPTION_PROMPT_MESSAGE = "Please write down your description:"

# TODO: We'll have to us a "plugin"/register design and possibly source this from the dbt project.
# !See issue #65 for most up to date thinking on this.
AVAILABLE_TESTS = ["unique", "not_null"]


class ConfirmQuestion(BaseModel):
    """Validation model for a question confirmation type payload."""

    type: str
    message: str
    name: str
    qmark: str = "?"
    auto_enter: bool = True

    @validator("type")
    def check_type_is_confirm(cls, value):
        assert value == "confirm", "'type' must be set to 'confirm' for this question type."
        return value


class ConfirmModelDoc(ConfirmQuestion):
    """Validatiorn model for a model documentation specific confirm question.

    Extends ConfirmQuestion to add stricter validation on the name field.
    """

    @validator("name")
    def check_name_is_wants_to_document(cls, value):
        assert (
            value == "wants_to_document_model"
        ), "'name' must be set to 'wants_to_document_model'."
        return value


class FreeTextInput(BaseModel):
    """Validation model for a free text input question payload."""

    type: str
    name: str
    message: str

    @validator("type")
    def check_type_is_text(cls, value):
        assert value == "text", "'type' must be set to 'text' for this question type."
        return value


class DescriptionTextInput(FreeTextInput):
    """Validation model for a model or column description text question payload.

    Extends FreeTextInput to add a stricter message validation.
    """

    message: str = DESCRIPTION_PROMPT_MESSAGE

    @validator("message")
    def check_message_text(cls, value):
        assert (
            value == DESCRIPTION_PROMPT_MESSAGE
        ), "Overriding question prompt is not allowed for a description question"
        return value


class MultipleChoiceInput(BaseModel):
    """Validation model for a multiple choice question payload."""

    type: str
    choices: List[str]
    name: str

    @validator("type")
    def check_type_is_choice(cls, value):
        assert value == "checkbox", "'type' must be set to checkbox."
        return value

    @validator("name")
    def check_name_is_cols_to_document(cls, value):
        assert value == "cols_to_document", "'name' must be set to 'cols_to_document'."
        return value


class MultipleChoiceInputWithDict(MultipleChoiceInput):
    """Validation model for a multiple choice question type where the choices list is a dict."""

    choices: Dict[str, str]  # type: ignore # because mypy doesn't like this but I'm ok with it.


class UserInputCollector:
    """User input collector class.

    Validates and orchestrates the interface between the dbt-sugar back end and the questionary API.
    Uses questionary API for CLI collection and imposes some stronger validation rules or flows.

    MAINTENANCE: When we add a lot more question types, we'll likely want to refactor this class to
    use a factory as the view code is going to get long and full of ifs. For now this should get us
    started in the right direction.
    """

    def __init__(
        self,
        question_type: str,
        question_payload: Sequence[Mapping[str, Any]],
        ask_for_tests: bool = True,
        ask_for_tags: bool = True,
        is_paginated: bool = False,
        is_first_page: Optional[bool] = None,
    ) -> None:
        """Constructor for UserInpurCollector.

        Expects a question type and a question payload. The payload must be a list of dictionaries,
        following the `questionary.prompt()` API.

        Args:
            question_type (str): name of the questioning flow type.
                Currently supported: ['model', 'undocumented_cols'].
            question_payload (List[Mapping[str, Any]]): List of dicts following the
                `questionary.prompt()` API requirements (examples below):
                    https://questionary.readthedocs.io/en/stable/pages/advanced.html#question-dictionaries

        Question Payload Examples:
        ```python
        model_doc_payload = [
            {
                "type": "confirm",
                "name": "wants_to_document_model",
                "message": "Model Description: This is my previously documented model. Document?",
                "default": True,
            },
            {
                "type": "text",
                "name": "model_description",
                "message": "Please write down your description:"
            },
        ]

        undocumented_columns_payload = [
            {
                "type": "checkbox",
                "name": "cols_to_document",
                "choices": ["col_a", "col_b"],
                "message": "Select the columns you want to document.",
            }
        ]
        ```
        """
        self._question_type = question_type
        self._question_payload = question_payload
        self._is_valid_question_payload = False
        self._ask_for_tests = ask_for_tests
        self._ask_for_tags = ask_for_tags
        if is_paginated:
            assert isinstance(
                is_first_page, bool
            ), "When using a paginated flow is_first_page cannot be None"
        self._is_paginated = is_paginated
        self._is_first_page = is_first_page

    def _validate_question_payload(self) -> None:
        assert isinstance(self._question_payload, list), "Question payload must be a list of dicts."
        if self._question_type == "model":
            assert (
                len(self._question_payload) == 2
            ), "Payload for model question must contain two dicts."
            for payload_element_index, payload_element in enumerate(self._question_payload):
                if payload_element_index == 0:
                    ConfirmModelDoc(**payload_element)
                elif payload_element_index == 1:
                    DescriptionTextInput(**payload_element)

        elif self._question_type == "undocumented_columns":
            MultipleChoiceInput(**self._question_payload[0])

        elif self._question_type == "documented_columns":
            MultipleChoiceInputWithDict(**self._question_payload[0])

        else:
            raise NotImplementedError(f"{self._question_type} is not implemented.")

        self._is_valid_question_payload = True

    def collect_rich_user_input(self) -> str:
        """Uses click to open up a text editor to collect rich user input.

        Returns:
            str: string version of the text input which will then be loaded.
        """
        tests = click.edit(extension=".yml")
        if tests:
            tests = tests.replace("\t", "    ")
            tests = yaml.safe_load(tests)
        return tests

    def _iterate_through_columns(
        self, cols: List[str]
    ) -> Mapping[str, Mapping[str, Union[str, List[str]]]]:
        """Iterates through a provided list of columns collects documentation info.

        More specifically it asks for the following
        - a column documentation
        - whether user wants to add tests
            - and if so which tests to add from a list of pre-defined/aggreed tests.

        Args:
            cols (List[str]): List of columns to collect description and tests info from.
            ask_for_tests (bool, optional): When set to true UI will ask for whether the user wants
            to add tests for each of the columns. Defaults to True.

        Returns:
            Mapping[str, Mapping[str, Union[str, List[str]]]]: API paylad to be passed to back-end
            that looks like this:
            ```python
            {
                'col_a': {
                    'description': 'Description for col a',
                    'tests': ['unique'],
                    'tags': ['PII']
                    },
                'col_b': {'description': 'Description for col b'}
            }
            ```
        """
        results = dict()
        for column in cols:
            description = questionary.text(
                message=f"Column: '{column}': {DESCRIPTION_PROMPT_MESSAGE}"
            ).unsafe_ask()

            if description:
                results.update({column: {"description": description}})
            else:
                # we still want to make the column end up in the dict but we don't put a desc if
                # empty... not sure about this
                results.update({column: {}})

            # kick in the test flows
            if self._ask_for_tests:
                wants_to_add_tests = questionary.confirm(
                    message="Would you like to add any tests?"
                ).unsafe_ask()
                if wants_to_add_tests:
                    wants_to_pop_editor = questionary.checkbox(
                        # message="Do you want to add a complex or custom tests? If, so we'll open a test editor for "
                        # f"you, otherwise you can choose from the following builtins: {AVAILABLE_TESTS}"
                        message="Do you want to add simple builtin tests or custom/more complex tests?"
                        "\nIf you choose 'complex tests' we'll take you to a text editor so you can add"
                        " those more easily.",
                        choices=["builtin tests", "complex tests"],
                    ).unsafe_ask()
                    if "complex tests" in wants_to_pop_editor:
                        tests = self.collect_rich_user_input()
                    else:
                        tests = questionary.checkbox(
                            message="Please select one or more tests from the list below",
                            choices=AVAILABLE_TESTS,
                        ).unsafe_ask()
                    if tests:
                        results[column]["tests"] = tests

            # kick in the tags flow
            if self._ask_for_tags:
                wants_to_add_tags = questionary.confirm(
                    message="Would you like to add any tags?"
                ).unsafe_ask()
                if wants_to_add_tags:
                    tags = questionary.text(
                        message="Provide a comma-separated list of tags"
                    ).unsafe_ask()
                    if tags:
                        tags = self.__split_comma_separated_str(tags)
                        results[column]["tags"] = tags
            # remove the column if no info has been given (no tests, and no description).
            if not results[column]:
                _ = results.pop(column)
        return results

    @staticmethod
    def __split_comma_separated_str(tags: str) -> List[str]:
        _tags = []
        if isinstance(tags, str):
            _tags = tags.split(",")
            _tags = [s.strip() for s in _tags]
            return _tags
        raise TypeError(f"Tags can only be strings. You provided a {type(tags)}")

    @staticmethod
    def _document_model(
        question_payload: Sequence[Mapping[str, Any]]
    ) -> Dict[str, Union[bool, str]]:
        results = dict()
        for i, payload_element in enumerate(question_payload):
            results.update(questionary.prompt(payload_element))
            collect_model_description = i < 1 and results.get("wants_to_document_model") is True

            if not results.get("model_description"):
                # we return an empty dict if user decided to not enter a description in the end.
                results = {}
            if not collect_model_description:
                # if the user doesnt want to document the model we exit early even if the payload
                # has a second entry which would trigger the description collection
                break
        return results

    def _set_quantifier_word(self) -> str:
        if self._is_first_page and self._is_paginated:
            return "first"
        if not self._is_paginated:
            return "model's"
        return "next"

    def _document_undocumented_cols(
        self,
        question_payload: Sequence[Mapping[str, Any]],
    ) -> Mapping[str, Mapping[str, Union[str, List[str]]]]:

        columns_to_document = question_payload[0].get("choices", list())
        quantifier_word = self._set_quantifier_word()
        # check if user wants to document all columns
        document_all_cols = questionary.confirm(
            message=(
                f"Do you want to document all of the {quantifier_word} {len(columns_to_document)} "
                "undocumented columns?"
            ),
            auto_enter=True,
        ).unsafe_ask()

        if document_all_cols:
            return self._iterate_through_columns(cols=columns_to_document)
        # get the list of columns from user
        columns_to_document = questionary.prompt(question_payload)
        return self._iterate_through_columns(
            cols=columns_to_document["cols_to_document"],
        )

    def _document_already_documented_cols(
        self,
        question_payload: Sequence[Mapping[str, Any]],
    ) -> Mapping[str, Mapping[str, Union[str, List[str]]]]:
        mutable_payload = copy.deepcopy(question_payload)
        mutable_payload = cast(Sequence[Dict[str, Any]], mutable_payload)

        # massage the question payload
        choices = [
            f"{col} | {desc}" for col, desc in mutable_payload[0].get("choices", dict()).items()
        ]

        mutable_payload[0].update({"choices": choices})

        # ask user if they want to see any of the documented columns?
        results = {}
        document_any_columns = questionary.confirm(
            message="Do you want to document any of the already documented columns in this model?",
            auto_enter=True,
        ).unsafe_ask()

        if document_any_columns:
            columns_to_document = questionary.prompt(mutable_payload)
            _results = self._iterate_through_columns(
                cols=columns_to_document["cols_to_document"],
            )

            # remove description from col key
            for col, desc in _results.items():
                stripped_col_name = col.split("|")[0].strip()
                results.update({stripped_col_name: desc})
            return results

        return results

    def collect(self) -> Mapping[str, Any]:
        """Question orchestractor function.

        Depending on the question type provided on the class will call payload validation and
        orchestrate documentation input collection.

        - When `question_type` == 'model' the following dict is returned:
        ```python
        {'wants_to_document_model': True, 'model_description': 'New def'}
        ```
        - When `question_type` == 'undocumented_columns` or `documented_columns` the following dict
        is returned:
        ```python
        {
            'col_a': {
                'description': 'Description for col a',
                'tests': ['unique']
                },
            'col_b': {'description': 'Description for col b'}
        }
        ```

        Raises:
            NotImplementedError: When a non supported `question_type` is provided

        Returns:
            Mapping[str, Union[bool, str]]: Response object for the backend documentation task.
                See above for examples.
        """
        self._validate_question_payload()

        # Model Documentation Flow
        if self._question_type == "model":
            return self._document_model(self._question_payload)

        # Undocumented Columns Documentation Flow
        if self._question_type == "undocumented_columns":
            return self._document_undocumented_cols(self._question_payload)

        if self._question_type == "documented_columns":
            return self._document_already_documented_cols(self._question_payload)

        raise NotImplementedError(f"{self._question_type} is not implemented.")



"""Bootstrap module. Generates placeholders for all models in a dbt project."""


import functools
import operator
import os
import re
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Sequence, Union

from dbt_sugar.core.clients.dbt import DbtProfile
from dbt_sugar.core.clients.yaml_helpers import open_yaml, save_yaml
from dbt_sugar.core.config.config import DbtSugarConfig
from dbt_sugar.core.flags import FlagParser
from dbt_sugar.core.task.doc import DocumentationTask


@dataclass
class DbtModelsDict:
    """Data class for dbt model info.

    We make it a dataclass instead of a dict because the types
    inside a dict would have been too messy and was upstetting type checkers.
    """

    model_name: str
    model_path: Path
    model_columns: Sequence[str]


class BootstrapTask(DocumentationTask):
    """Sets up methods and orchestration of the bootstrap task.

    The bootstrap task is a task that iterates through all the models
    in a dbt project, checks the tables exist on the db, and generates
    placeholder model descriptor files (schema.yml) for any column or models
    that have not yet been documented.
    """

    def __init__(
        self,
        flags: FlagParser,
        dbt_path: Path,
        sugar_config: DbtSugarConfig,
        dbt_profile: DbtProfile,
    ) -> None:
        # we specifically run the super init because we need to populate the cache
        # of all dbt models, where they live etc
        super().__init__(
            flags=flags, dbt_profile=dbt_profile, config=sugar_config, dbt_path=dbt_path
        )
        self.dbt_models_dict: Dict[str, Union[Path, List[str]]] = {}
        self._dbt_profile = dbt_profile
        self.schema = self._dbt_profile.profile.get("target_schema", "")

        self.dbt_models_data: List[DbtModelsDict] = []

    def build_all_models_dict(self) -> None:
        """Walk through all .sql files and load their info (name, path etc) into a dict."""
        _dbt_models_data = []
        for root, _, files in os.walk(self.repository_path):
            if not re.search(self._excluded_folders_from_search_pattern, root):
                _dbt_models_data.append(
                    [
                        DbtModelsDict(
                            model_name=f.replace(".sql", ""),
                            model_path=Path(root, f),
                            model_columns=[],
                        )
                        for f in sorted(files)
                        if f.lower().endswith(".sql")
                        and f.lower().replace(".sql", "")
                        not in self._sugar_config.dbt_project_info.get("excluded_models", [])
                    ]
                )
                self.dbt_models_data = functools.reduce(operator.iconcat, _dbt_models_data, [])

    def add_or_update_model_descriptor_placeholders(self, is_test: bool = False):
        connector = self.get_connector()
        for model_info in self.dbt_models_data:
            model_descriptor_content = {}
            model_info.model_columns = connector.get_columns_from_table(
                model_info.model_name,
                self.schema,
                use_describe=self._sugar_config.dbt_project_info.get(
                    "use_describe_snowflake", False
                ),
            )
            (
                model_descriptor_path,
                descriptor_file_exists,
                is_already_documented,
            ) = self.find_model_schema_file(model_name=model_info.model_name)
            if descriptor_file_exists and model_descriptor_path:
                model_descriptor_content = open_yaml(
                    model_descriptor_path,
                    preserve_yaml_order=self._sugar_config.config.get("preserve_yaml_order", False),
                )

            model_descriptor_content = self.create_or_update_model_entry(
                is_already_documented,
                model_descriptor_content,
                model_name=model_info.model_name,
                columns_sql=model_info.model_columns,
            )
            if is_test:
                return self.order_schema_yml(model_descriptor_content)
            if model_descriptor_path:
                if self._sugar_config.config.get("preserve_yaml_order"):
                    save_yaml(
                        model_descriptor_path, model_descriptor_content, preserve_yaml_order=True
                    )
                else:
                    save_yaml(
                        model_descriptor_path, self.order_schema_yml(model_descriptor_content)
                    )

    def run(self) -> int:
        self.build_all_models_dict()
        self.add_or_update_model_descriptor_placeholders()
        return 0



"""
Module base connector.

Only use this class implemented by a child connector.
"""
from abc import ABC
from typing import Dict, Sequence

import sqlalchemy


class BaseConnector(ABC):
    """
    Class base connector.

    Parent class of all the connectors.
    """

    def __init__(
        self,
        connection_params: Dict[str, str],
    ) -> None:
        """
        Creates the URL and the Engine for future connections.

        Args:
            connection_params (Dict[str, str]): Dict containing database connection
                parameters and credentials.
        """
        self.engine = sqlalchemy.create_engine(**connection_params)

    def get_columns_from_table(
        self, target_table: str, target_schema: str, use_describe: bool = False
    ) -> Sequence[str]:
        """
        Method that creates cursor to run a query.

        Args:
            target_table (str): table to get the columns from.
            target_schema (str): schema to get the table from.

        Returns:
            Optional[Sequence[str]]: With the results of the query.
        """
        inspector = sqlalchemy.engine.reflection.Inspector.from_engine(self.engine)
        columns = inspector.get_columns(target_table, target_schema)
        return [column["name"] for column in columns]

    def run_test(self, test_name: str, schema: str, table: str, column: str) -> bool:
        """
        Method to run pre-defined tests before we add them to the schema.yaml.

        We defined the test dictionary and call the test that the user wants.

        Args:
            test_name(str): Name of the test to run. (For now only "unique" and "not null" are supported).
            schema (str): Name of the schema in which the table to be tested lives.
            table (str): Name of the table to on which to run the test.
            column (str): Name of the column on which to run the test.
        Returns:
            boolean: True if the test is passes, and False if it fails.
        """
        TESTS = {
            "unique": f"""select count(*) as errors from(
                select {column} from {schema}.{table} where {column} is not null group by {column} having count(*) > 1 )
                errors""",
            "not_null": f"select count(*) as errors from {schema}.{table} where {column} is null",
        }
        query = TESTS[test_name]
        return self.execute_and_check(query)

    def execute_and_check(self, query) -> bool:
        """
        Method to run a test query and check test results.

        Args:
            query(str): SQL query string to execute.
        Returns:
            boolean: True if the test passes, and False if it fails.
        """
        with self.engine.connect() as cursor:
            result = cursor.execute(query).fetchone()
            if result[0] < 1:
                return True
        return False
